/*
******************************************************************************

   BLT implementation for X target - API functions.

   Copyright (C) 2001	Christoph Egger	[Christoph_Egger@t-online.de]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <math.h>

#include "xblt.h"

#include <ggi/display/xcommon.h>
#include <ggi/display/x.h>

#include <X11/X.h>

#include <ggi/ggi.h>




int BLT_X_tie(ggiBlt_t blt, ggiGA_resource_handle handle)
{
	struct xblt_bob *bob = NULL;
	int rc = GALLOC_OK;

	if (blt == NULL) return GGI_EARGINVAL;
	if (handle == NULL) return GGI_EARGINVAL;

	bob = malloc(sizeof(struct xblt_bob));
	if (!bob) return GGI_ENOMEM;
	blt->target = (void *)bob;

	bob->bob = ggiGAGetPriv(handle);
	LIB_ASSERT(bob->bob != NULL);
	if (!bob->bob) {
		free(bob);
		rc = GALLOC_EFAILED;
	}	/* if*/

	return rc;
}	/* BLT_X_tie */



int BLT_X_setimage(ggiBlt_t blt, ggi_visual_t vis,
		int x, int y, unsigned char *alpha)
{
	int rc = GALLOC_OK;
	ggiGA_resource_handle handle;
	ggi_xwin_common *priv;
	void *backgr = NULL;

	struct GA_x_bob *bob = NULL;

	priv = LIBGGI_PRIVATE(blt->vis);


	if (blt == NULL) return GGI_EARGINVAL;
	if (vis == NULL) return GGI_EARGINVAL;


	handle = blt->res_handle;

	/* This function is the special case, wher we must ignore
	 * GA_STORAGE_READONLY. Otherwise, we don't get, what we want.
	 */
	if (ggiGAIsShared(handle) && ggiGAHasCopyOnWrite(handle)) {
		rc = ggiGACopy2Write(&blt->res_handle);
		if (rc != GALLOC_OK) goto err0;
	}	/* if */



	backgr = malloc(blt->w * blt->h * sizeof(ggi_pixel));
	if (!backgr) goto err0;

	bob = GETBOB(blt);

	LIB_ASSERT(bob != NULL);
	LIB_ASSERT(bob->bobimage != NULL);


	if ((alpha != NULL) && (blt->alpha == NULL)) {
		blt->alpha = malloc(blt->w * blt->h *
				sizeof(unsigned char));
		if (!blt->alpha) goto err1;
	}	/* if */

	if ((alpha == NULL) && (blt->alpha != NULL)) {
		free(blt->alpha);
		blt->alpha = NULL;
	}	/* if */

	if (alpha != NULL) {
		memcpy(blt->alpha, alpha,
			blt->w * blt->h * sizeof(unsigned char));
	}	/* if */

	ggiGetBox(blt->vis, 0,0, blt->w,blt->h, backgr);
	ggiCrossBlit(vis, x,y, blt->w, blt->h, blt->vis, 0,0);

	ggiGetBox(blt->vis, 0,0, blt->w,blt->h, bob->bobimage);
	ggiPutBox(blt->vis, 0,0, blt->w,blt->h, backgr);

	ggiFlush(blt->vis);
	free(backgr);
	backgr = NULL;

	return GALLOC_OK;


err1:
	free(backgr);
err0:
	return rc;
}	/* BLT_X_setimage */









int BLT_X_crossblit(ggiBlt_t src, ggiBlt_t dst,
		    int width, int height,
		    enum ggiGA_rop rop)
{
	int rc = GALLOC_OK;
	ggiGA_resource_handle handle;
	int stretch;


	if (src == NULL) return GGI_EARGINVAL;
	if (dst == NULL) return GGI_EARGINVAL;

	handle = dst->res_handle;
	if (ggiGAIsReadOnly(handle)) {
		rc = GALLOC_EFAILED;
		goto exit;
	}	/* if */

	if (ggiGAIsShared(handle) && ggiGAHasCopyOnWrite(handle)) {
		rc = ggiGACopy2Write(&dst->res_handle);
		if (rc != GALLOC_OK) goto exit;
	}	/* if */

	/* Check if we perform stretching */
	if ((width == 0 && height == 0)
	  || (width == src->w && height == src->h)) {
		stretch = 0;	/* no stretching */
	} else {
		stretch = 1;	/* yes, we stretch */
	}	/* if */


	if (stretch) {
		rc = stretch_crossblit(src, dst, width, height, rop);
		goto exit;
	}	/* if */

	/* no stretching */
	if (rop != ROP_NOP) {
		rc = rop_crossblit(src, dst, rop);
		goto exit;
	}	/* if */

	/* do fast blitting here */
	rc = fast_crossblit(src, dst);
	goto exit;

exit:
	printf("%s:%s:%i: Check-POINT E\n", DEBUG_INFO);
	return rc;
}	/* BLT_X_crossblit */


int BLT_X_multicrossblit(ggiBlt_source_list src, size_t nmemb,
			 ggiBlt_t blt, int width, int height)
{
	int rc = GALLOC_OK;
	ggiGA_resource_handle handle;
	int stretch;


	if (src == NULL) return GGI_EARGINVAL;
	if (blt == NULL) return GGI_EARGINVAL;


	handle = blt->res_handle;
	if (ggiGAIsReadOnly(handle)) {
		rc = GALLOC_EFAILED;
		goto exit;
	}	/* if */

	if (ggiGAIsShared(handle) && ggiGAHasCopyOnWrite(handle)) {
		rc = ggiGACopy2Write(&blt->res_handle);
		if (rc != GALLOC_OK) goto exit;
	}	/* if */


	/* Check if we perform stretching */
	if (width == 0 && height == 0) {
		stretch = 0;	/* no stretching */
	} else {
		stretch = 1;	/* yes, we stretch */
	}	/* if */


	if (stretch) {
		rc = stretch_multicrossblit(src, nmemb, blt, width, height);
		goto exit;
	}	/* if */


	/* do blitting using ROPs here */
	rc = rop_multicrossblit(src, nmemb, blt);
	goto exit;

exit:
	printf("%s:%s:%i: Check-POINT E\n", DEBUG_INFO);
	return rc;
}	/* BLT_X_multicrossblit */


int BLT_X_polycrossblit(ggiBlt_source_list src, size_t nmemb,
			ggiBlt_t blt, int width, int height)
{
	int rc = GALLOC_OK;
	ggiGA_resource_handle handle;
	int stretch;


	if (src == NULL) return GGI_EARGINVAL;
	if (blt == NULL) return GGI_EARGINVAL;


	handle = blt->res_handle;
	if (ggiGAIsReadOnly(handle)) {
		rc = GALLOC_EFAILED;
		goto exit;
	}	/* if */

	if (ggiGAIsShared(handle) && ggiGAHasCopyOnWrite(handle)) {
		rc = ggiGACopy2Write(&blt->res_handle);
		if (rc != GALLOC_OK) goto exit;
	}	/* if */


	/* Check if we perform stretching */
	if (width == 0 && height == 0) {
		stretch = 0;	/* no stretching */
	} else {
		stretch = 1;	/* yes, we stretch */
	}	/* if */


	if (stretch) {
		rc = stretch_polycrossblit(src, nmemb, blt, width, height);
		goto exit;
	}	/* if */


	/* do blitting using ROPs here */
	rc = rop_polycrossblit(src, nmemb, blt);
	goto exit;

exit:
	printf("%s:%s:%i: Check-POINT E\n", DEBUG_INFO);
	return rc;
}	/* BLT_X_polycrossblit */










int BLT_X_crossvis2vis(ggi_visual_t src, int sx, int sy, int sw, int sh,
		       ggi_visual_t dst, int dx, int dy, int width, int height,
		       enum ggiGA_rop rop, enum ggiGA_coordbase cm)
{
	int rc = GALLOC_OK;
	int stretch;


	if (src == NULL) return GGI_EARGINVAL;
	if (dst == NULL) return GGI_EARGINVAL;


	/* Check if we perform stretching */
	if ((width == 0 && height == 0)
	  || (width == sw && height == sh)) {
		stretch = 0;	/* no stretching */
	} else {
		stretch = 1;	/* yes, we stretch */
	}	/* if */


	if (stretch) {
		rc = stretch_crossvis2vis(src, sx, sy, sw, sh,
					dst, dx, dy, width, height,
					rop, cm);
		goto exit;
	}	/* if */


	/* no stretching */
	if (rop != ROP_NOP) {
		rc = rop_crossvis2vis(src, sx, sy, sw, sh,
					dst, dx, dy, rop, cm);
		goto exit;
	}	/* if */

	/* do fast blitting here */
	rc = fast_crossvis2vis(src, sx, sy, sw, sh, dst, dx, dy, cm);
	goto exit;

exit:
	printf("%s:%s:%i: Check-POINT E\n", DEBUG_INFO);
	return rc;
}	/* BLT_X_crossvis2vis */


int BLT_X_crossblit2vis(ggiBlt_t blt, ggi_visual_t vis,
			int dx, int dy, int width, int height,
			enum ggiGA_rop rop, enum ggiGA_coordbase cm)
{
	int rc = GALLOC_OK;
	int stretch;


	if (blt == NULL) return GGI_EARGINVAL;
	if (vis == NULL) return GGI_EARGINVAL;


	/* Check if we perform stretching */
	if ((width == 0 && height == 0)
	  || (width == blt->w && height == blt->h)) {
		stretch = 0;	/* no stretching */
	} else {
		stretch = 1;	/* yes, we stretch */
	}	/* if */


	if (stretch) {
		rc = stretch_crossblit2vis(blt, vis,
				dx, dy, width, height, rop, cm);
		goto exit;
	}	/* if */


	/* no stretching */
	if (rop != ROP_NOP) {
		rc = rop_crossblit2vis(blt, vis, dx, dy, rop, cm);
		goto exit;
	}	/* if */

	/* do fast blitting here */
	rc = fast_crossblit2vis(blt, vis, dx, dy, cm);
	goto exit;

exit:
	printf("%s:%s:%i: Check-POINT E\n", DEBUG_INFO);
	return rc;
}	/* BLT_X_crossblit2vis */


int BLT_X_blit2vis(ggiBlt_t blt, int dx, int dy, int width, int height,
		   enum ggiGA_rop rop)
{
	int rc = GALLOC_OK;
	int stretch;


	if (blt == NULL) return GGI_EARGINVAL;


	/* Check if we perform stretching */
	if ((width == 0 && height == 0)
	  || (width == blt->w && height == blt->h)) {
		stretch = 0;	/* no stretching */
	} else {
		stretch = 1;	/* yes, we stretch */
	}	/* if */


	if (stretch) {
		rc = stretch_blit2vis(blt, dx, dy, width, height, rop);
		goto exit;
	}	/* if */


	/* no stretching */
	if (rop != ROP_NOP) {
		rc = rop_blit2vis(blt, dx, dy, rop);
		goto exit;
	}	/* if */


	/* do fast blitting here */
	rc = fast_blit2vis(blt, dx, dy);

exit:
	return rc;
}	/* BLT_X_blit2vis */
